# -*- coding: utf-8 -*-
from dataclasses import dataclass
from pathlib import Path
from typing import Optional
from majordome.piv.plot import plot_histogram
from majordome.piv.plot import plot_images
from majordome.utilities import progress_bar
from openpiv.tools import imread
from openpiv.tools import save as txtsave
from openpiv.tools import transform_coordinates
from openpiv.pyprocess import extended_search_area_piv
from openpiv.pyprocess import get_coordinates
from openpiv.validation import sig2noise_val
from openpiv.filters import replace_outliers
from openpiv.scaling import uniform
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt


@dataclass
class PivConfig:
    """ Configuration to process files with OpenPIV.

    Attributes
    ----------
    time_step: float
        Time-step between consecutive images in seconds.
    scaling_factor: float
        Scaling factor from pixel to desired metric.
    window_size: int = 32
        Interrogation window size in pixels.
    search_area_size: int = 40
        Search windows size in pixels.
    overlap: int = 16
        Overlap between sliding windows in search.
    correlation_method: str = "linear"
        Correlation method (linear or circular).
    subpixel_method: str = "parabolic"
        Subpixel method (centroid, gaussian or parabolic).
    sig2noise_method: str = "peak2mean"
        Signal extraction method (peak2peak or peak2mean).
    peak_width_ignore: int = 1
        Width to ignore around peak when using `peak2peak` method.
    percentile: float = 5.0
        Percentile of data to be considered as outliers.
    min_sr: float = 1.05
        Minimum S/R to be considered valid data. Notice that if the
        threshold computed with `percentile` is below this value it
        is ignore and `min_sr` is used instead.
    outlier_method: str = "localmean"
        Outlier filter method (localmean, disk or distance).
    outlier_kernel_size: int = 2
        Outlier filter kernel size.
    outlier_max_iter: int = 100
        Outlier filter maximum iterations.
    outlier_tol: float = 0.0001
        Outlier filter tolerance.
    arrow_width: float = 0.003
        Width of arrows to draw over plot.
    arrow_scale: float = 300_000
        Multiplier to provide scaling to the arrows over plot.
    """
    time_step: float
    scaling_factor: float
    window_size: int = 32
    search_area_size: int = 40
    overlap: int = 16
    correlation_method: str = "linear"
    subpixel_method: str = "parabolic"
    sig2noise_method: str = "peak2mean"
    peak_width_ignore: int = 1
    percentile: float = 5.0
    min_sr: float = 1.05
    outlier_method: str = "localmean"
    outlier_kernel_size: int = 2
    outlier_max_iter: int = 100
    outlier_tol: float = 0.0001
    arrow_width: float = 0.003
    arrow_scale: float = 300_000

    def threshold(self, s2n: np.ndarray):
        """ Compute signal-to-noise threshold to drop outliers. """
        return max(self.min_sr, np.percentile(s2n, self.percentile))


def piv_workflow(
    conf: PivConfig,
    images: list[str | Path],
    output_dir: str | Path,
    every: Optional[int] = 1
) -> pd.DataFrame:
    """ Process a batch of sequential images provided in list.

    Parameters
    ----------
    conf: PivConfig
        Configuration object with OpenPIV parameters.
    images: list[PathLike]
        List of sorted images to run in batch mode.
    output_dir: PathLike
        Path to directory to generate output files.
    every: Optional[int] = 1
        Step between consecutive files in list to compute velocities.

    Returns
    -------
    pd.DataFrame
        Data frame with time, mean velocities and standard deviations.
    """
    df = pd.DataFrame(columns=["t", "u_mean", "u_std", "v_mean",
                               "v_std",  "s_mean", "s_std"])
    names = ["x", "y", "u", "v", "mask"]

    pairs = zip(images[:-every], images[every:])
    pbar = progress_bar(pairs, size=len(images)-1, enum=True)

    for count, (file_a, file_b) in pbar:
        savehist = output_dir / f"transition_{count:04d}_hist.png"
        saveimgs = output_dir / f"transition_{count:04d}_imgs.png"
        datafile = output_dir / f"transition_{count:04d}_data.dat"

        _pair_process(conf, file_a, file_b, datafile,
                      savehist=savehist, saveimgs=saveimgs,
                      counter=count)

        data = pd.read_csv(datafile, sep="\t", skiprows=1, names=names)
        data["s"] = np.sqrt(pow(data["u"], 2) + pow(data["v"], 2))

        t = (count / every) * conf.time_step
        row = data.describe().T[["mean", "std"]].loc[["u", "v", "s"]]
        df.loc[count] = t, *row.to_numpy().flatten()

    df.to_csv(output_dir / "results.csv", index=False)

    return df


def _pair_process(
    conf: PivConfig,
    file_a: str | Path,
    file_b: str | Path,
    datafile: str | Path,
    savehist: Optional[str | Path] = None,
    saveimgs: Optional[str | Path] = None,
    counter: Optional[int] = 0,
    every: Optional[int] = 1,
    figsize: Optional[tuple[float, float]] = (12, 12),
    dpi: Optional[int] = 150
) -> None:
    """ Process a pair of images and generates velocity table.

    Parameters
    ----------
    conf: PivConfig
        Configuration object with OpenPIV parameters.
    file_a: PathLike
        Path to first file in the frame series.
    file_b: PathLike
        Path to second file in the frame series.
    datafile: PathLike
        Path to dump extracted data file.
    savehist: Optional[PathLike] = None
        If provided, save S/N ratio histogram to this file.
    saveimgs: Optional[PathLike] = None
        If provided, save sequence of images to this file.
    counter: Optional[int] = 0
        If provided, index first image in the series with this value.
    every: Optional[int] = 1
        Step between consecutive files in list to compute velocities.
    figsize: Optional[FigSize] = (12, 12)
        Figure size to be provided to `matplotlib.
    dpi: Optional[int] = 150
        Figure pixel density to be provided to `matplotlib.
    """
    frame_a = imread(file_a).astype(np.int32)
    frame_b = imread(file_b).astype(np.int32)

    u, v, s2n = extended_search_area_piv(
        frame_a, frame_b,
        window_size=conf.window_size,
        overlap=conf.overlap,
        dt=conf.time_step,
        correlation_method=conf.correlation_method,
        subpixel_method=conf.subpixel_method,
        sig2noise_method=conf.sig2noise_method,
        width=conf.peak_width_ignore,
        search_area_size=conf.search_area_size,
        normalized_correlation=True
    )

    threshold = conf.threshold(s2n)

    u, v, mask = sig2noise_val(u, v, s2n=s2n, w=None,
                               threshold=threshold)

    u, v = replace_outliers(u, v, w=None,
                            method=conf.outlier_method,
                            max_iter=conf.outlier_max_iter,
                            tol=conf.outlier_tol,
                            kernel_size=conf.outlier_kernel_size)

    x, y = get_coordinates(image_size=frame_a.shape,
                           search_area_size=conf.search_area_size,
                           overlap=conf.overlap)

    x, y, u, v = uniform(x, y, u, v, scaling_factor=conf.scaling_factor)
    x, y, u, v = transform_coordinates(x, y, u, v)

    txtsave(x, y, u, v, mask, datafile)

    if savehist is not None:
        plot_histogram(s2n, threshold).savefig(savehist, dpi=dpi)

    if saveimgs is not None:
        plot_images(frame_a, frame_b, datafile, file_a, counter, every,
                    conf.scaling_factor, conf.window_size,
                    conf.arrow_scale, conf.arrow_width,
                    figsize=figsize).savefig(saveimgs, dpi=dpi)

    plt.close("all")
