# -*- coding: utf-8 -*-
from pathlib import Path
from typing import Optional
from imageio.core.util import Array
from matplotlib.figure import Figure
from openpiv.tools import imread
from openpiv.tools import negative
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt


def plot_mean_speed(
    df: pd.DataFrame,
    figsize: Optional[tuple[float, float]] = (6, 5),
    add_bounds: Optional[bool] = True
) -> Figure:
    """ Plot mean measured speed over time from table.

    This function expects data frame as structured by an implementation
    of `piv_workflow`, *i.e.* with header names from that function.

    Parameters
    ----------
    df: pd.DataFrame
        Data frame as exported by `piv_workflow` implementation.
    figsize: Optional[FigSize] = (6, 5)
        Figure size to be provided to `matplotlib.
    add_bounds: Optional[bool] = True
        If `True`, display mean plus/minus one standard deviation.

    Returns
    -------
    Figure
        A `matplotlib` figure for further manipulation/dumping.
    """
    t = 1000 * df["t"].to_numpy()
    s_mean = df["s_mean"].to_numpy()
    s_std = df["s_std"].to_numpy()

    plt.close("all")
    plt.style.use("seaborn-white")
    fig = plt.figure(figsize=figsize)

    plt.plot(t, s_mean + 0 * s_std, label="$\\mu$")

    if add_bounds:
        plt.plot(t, s_mean + 1 * s_std, label="$\\mu + \\sigma$")
        plt.plot(t, s_mean - 1 * s_std, label="$\\mu - \\sigma$")

    plt.grid(linestyle=":")
    plt.xlabel("Time [ms]")
    plt.ylabel("Projected speed [mm/s]")
    plt.legend(loc=1, frameon=True)
    plt.tight_layout()

    return fig


def plot_histogram(
    s2n: np.ndarray,
    threshold: float,
    bins: Optional[int] = 50,
    figsize: Optional[tuple[float, float]] = (6, 5)
) -> Figure:
    """ Display signal-to-noise ratio histogram.

    Parameters
    ----------
    s2n: np.ndarray
        Array 2D of signal-to-noise values to plot distribution.
    threshold: float
        Threshold value to display point of cut-off over histogram.
    bins: Optional[int] = 50
        Number of bins to divide data over in histogram.
    figsize: Optional[FigSize] = (6, 5)
        Figure size to be provided to `matplotlib.

    Returns
    -------
    Figure
        A `matplotlib` figure for further manipulation/dumping.
    """
    plt.close("all")
    plt.style.use("seaborn-white")
    fig, ax = plt.subplots(figsize=figsize)

    ax.hist(s2n.flatten(), bins=bins)
    ax.axvline(threshold, color="k")
    ax.grid(linestyle=":")
    ax.set_xlabel("Ratio S/N")
    ax.set_ylabel("Count")

    fig.tight_layout()
    return fig


def plot_images(
    frame_a: Array,
    frame_b: Array,
    txtfile: str | Path,
    file_a: str | Path,
    counter: int,
    every: int,
    scaling_factor: float,
    window_size: int,
    arrow_scale: float,
    arrow_width: float,
    figsize: Optional[tuple[float, float]] = (8, 8)
) -> Figure:
    """ Display frames at different stages of development.

    Parameters
    ----------
    frame_a: Array
        First frame in the series.
    frame_b: Array
        Second frame in the series.
    txtfile: PathLike
        Path to data file with velocity profiles.
    file_a: PathLike
        Path to image file originating `frame_a`.
    counter: int
        Step number used to provide titles to the consecutive images.
    every: int
        Step between consecutive files in list to compute velocities.
    scaling_factor: float
        Conversion factor from pixels to millimeters.
    window_size: int
        Window size used during velocity profile extraction.
    arrow_scale: float
        Multiplier to provide scaling to the arrows over plot.
    arrow_width: float
        Width of arrows to draw over plot.
    figsize: Optional[FigSize] = (8, 8)
        Figure size to be provided to `matplotlib.

    Returns
    -------
    Figure
        A `matplotlib` figure for further manipulation/dumping.
    """
    # Compute difference between frames for speed intuitive view.
    frame_c = pow(frame_a, 2) - pow(frame_b, 2)

    plt.close("all")
    plt.style.use("seaborn-white")

    fig, ax = plt.subplots(2, 2, figsize=figsize)

    ax[0, 0].imshow(frame_a, cmap=plt.cm.hot)
    ax[0, 1].imshow(frame_b, cmap=plt.cm.hot)
    ax[1, 0].imshow(frame_c, cmap=plt.cm.hot)

    _display_vector_field(txtfile, window_size, ax[1, 1], image_name=file_a,
                          scaling_factor=scaling_factor, scale=arrow_scale,
                          width=arrow_width)

    ax[0, 0].grid(linestyle=":", color="k")
    ax[0, 1].grid(linestyle=":", color="k")
    ax[1, 0].grid(linestyle=":", color="k")
    ax[1, 1].grid(linestyle=":", color="k")

    ax[0, 0].set_title(f"Frame {counter + 0}")
    ax[0, 1].set_title(f"Frame {counter + every}")
    ax[1, 0].set_title("Movement shadows")
    ax[1, 1].set_title("Detected vectors")

    fig.tight_layout()

    return fig


def _display_vector_field(
        filename: str | Path, 
        window_size: int,
        ax: object,
        image_name: Optional[str | Path] = None,
        scaling_factor: Optional[int] = 1,
        width: Optional[float] = 0.0025, 
        scale: Optional[float] = 1.0
    ):
    """ Alternative to `openpiv.tools.display_vector_field`. """
    a = np.loadtxt(filename)
    x, y, u, v, mask = a[:, 0], a[:, 1], a[:, 2], a[:, 3], a[:, 4]

    if image_name is not None:
        im = negative(imread(image_name))
        xmax = np.amax(x) + window_size / (2 * scaling_factor)
        ymax = np.amax(y) + window_size / (2 * scaling_factor)
        ax.imshow(im, cmap="Greys_r", extent=[0.0, xmax, 0.0, ymax])

    invalid = mask.astype("bool")
    valid = ~invalid

    xv, yv, xi, yi = x[valid], y[valid], x[invalid], y[invalid]
    uv, vv, ui, vi = u[valid], v[valid], u[invalid], v[invalid]

    # TODO add alternative color scale based on sqrt(u**2 + v**2)!
    ax.quiver(xi, yi, ui, vi, color="r", width=width, scale=scale)
    ax.quiver(xv, yv, uv, vv, color="b", width=width, scale=scale)
