from typing import Optional
from majordome.units.flow_units import FlowUnits
from majordome.simulate.combustion.heating_value import HydrocarbonHeatingValue
import cantera as ct
import numpy as np


class BurnerFlowRatesCalculator:
    """ Compute nominal flow rates to match equivalence ratio.
    
    Parameters
    ----------
    mech : str
        Path to Cantera kinetics mechanism to be used.
    """
    def __init__(self, mech: str) -> None:
        self._mech = mech
        self._fuel = ct.Solution(self._mech)
        self._oxid = ct.Solution(self._mech)
        self._mixt = ct.Solution(self._mech)

        # Normal concentration in [kmol/Nm³] to [mol/Nm³]
        self._normal_conc = FlowUnits().normal_concentration
        self._normal_conc *= 1000

    def _get_molar_amounts(self) -> list[float]:
        """ Compute mole fractions of fuel and oxidizer. """
        Af = np.array([*self._fuel.X, 1])
        Ao = np.array([*self._oxid.X, 1])
        Am = np.array([*self._mixt.X, 1])

        mwf = self._fuel.mean_molecular_weight
        mwo = self._oxid.mean_molecular_weight
        mwm = self._mixt.mean_molecular_weight

        M = np.vstack((Af, Ao)).reshape((2, -1)).T
        x, res1, _, _ = np.linalg.lstsq(M, Am, rcond=None)
        res2 = np.dot(x, [mwf, mwo])

        # Arbitrary convergence check.
        if not (res1 < 1.0e-10) or not np.isclose(res2, mwm):
            raise ValueError(
                f"Problem formulation is wrong: "
                f"Residual of LSQ = {res1} and "
                f"residual = {res2}"
            )

        return x.flatten()

    def _get_fuel_heating_value(self, method: str) -> float:
        """ Compute fuel heating value with provided method. """
        gas = ct.Solution(self._mech)
        hcv = HydrocarbonHeatingValue(gas)
        lhv, hhv = hcv.heating_values(
            self._fuel.mole_fraction_dict(), 
            self._oxid.mole_fraction_dict()
        )

        match method.upper():
            case "HHV":
                hv = hhv
            case "LHV":
                hv = lhv
            case _:
                raise ValueError(f"Unknown method: {method}")

        # Here rho is `kg/Nm³`! Use MIXTURE molar mass!
        rho = self._normal_conc * self._mixt.mean_molecular_weight / 1000

        # `cap` is expected in `kWh/Nm³`.
        cap = hv / (3600 * rho)
        print(f"Corrected {method} = {cap:.1f} kWh/Nm³")

        return cap

    def _ndot_fuel(self, hdot: float, cap: float) -> tuple[float, float]:
        """ Compute fuel molar flow rate to reach energy level. """
        ndot = (self._normal_conc * hdot / cap) / 3600.0
        return ndot

    def mole_to_mass_flow(self, fuel_ndot, oxid_ndot):
        """ Convert molar units to mass units flow rates. """
        fuel_mdot = fuel_ndot * self._fuel.mean_molecular_weight / 1000
        oxid_mdot = oxid_ndot * self._oxid.mean_molecular_weight / 1000
        return fuel_mdot, oxid_mdot

    def set_states(self,
            phi,
            fuel,
            oxid,
            T=300, 
            P=ct.one_atm,
            basis="mole"
        ):
        """ Set states of fuel, oxidizer, and target mixture.
        
        Parameters
        ----------
        phi: float
            Feed equivalence ratio used to set initial composition.
        fuel: dict[str, float]
            Dictionary of fuel composition compatible with Cantera.
        oxid: dict[str, float]
            Dictionary of oxidizer composition compatible with Cantera.
        T: Optional[float] = 300.0
            Feed gas temperature in kelvin.
        P: Optional[float] = ct.one_atm
            Feed gas pressure in pascal.
        basis: Optional[str] = "mole"
            Whether composition is in `mole` or `mass` fracitons.
        """
        self._fuel.TPX = T, P, fuel
        self._oxid.TPX = T, P, oxid

        self._mixt.TP = T, P
        self._mixt.set_equivalence_ratio(phi, fuel, oxid, basis=basis)

    def get_mole_flow_rates(self,
            hdot: float,
            method: Optional[str] = "HHV",
            cap: Optional[float] = None
        ) -> tuple[float, float]:
        """ Compute required molar flow rates for equivalence ratio. 
        
        Parameters
        ----------
        hdot: float
            Required energy output in [kW] in `cap` basis.
        method: Optional[str] = "HHV"
            Method to evaluate heating value, HHV or LHV.
        cap: Optional[float] = None
            Fuel energy release capacity in [kWh/Nm³]. This entry is
            ignored if a method is provided.

        Returns
        -------
        tuple[float, float]
            Fuel and oxidized flow rates in [mol/s].
        """
        if cap is None or method is not None:
            cap = self._get_fuel_heating_value(method)

        x = self._get_molar_amounts()
        fuel_ndot = self._ndot_fuel(hdot, cap)
        oxid_ndot = fuel_ndot * x[1] / x[0]
        return fuel_ndot, oxid_ndot

    def get_mole_flow_from_fuel_flow(self,
            fuel_vdot: float
        ) -> tuple[float, float]:
        """ Get flow with same stoichiometry and imposed volume flow.
        
        Parameters
        ----------
        fuel_vdot: float
            Required fuel flow rate in [Nm³/h].

        Returns
        -------
        tuple[float, float]
            Fuel and oxidized flow rates in [mol/s].
        """
        x = self._get_molar_amounts()
        fuel_ndot = self._normal_conc * fuel_vdot / 3600.0
        oxid_ndot = fuel_ndot * x[1] / x[0]
        return fuel_ndot, oxid_ndot

