# -*- coding: utf-8 -*-
from typing import Optional
from matplotlib.figure import Figure
import cantera as ct
import matplotlib.pyplot as plt


class CombustorPSR:
    """ Single chamber perfect-stirred reactor combustor.
    
    This class is based on this Cantera example script:
    - cantera.org/examples/python/reactors/combustor.py.html

    It is a practical tool for comparing/evaluating different
    combustion mechanisms and their validity for an application.

    Parameters
    ----------
    mech : str
        Path to Cantera kinetics mechanism to be used.
    """
    def __init__(self, mech: str) -> None:
        self._gas = ct.Solution(mech)
        self._inlet = None
        self._combustor = None
        self._exhaust = None
        self._tau = None
        self._sim = None
        self._states = None

    def _equilibrate(self):
        """ Initialize reactor composition with equilibrium state. """
        self._gas.equilibrate("HP")
        self._combustor = ct.IdealGasReactor(self._gas, volume=1.0)
        self._exhaust = ct.Reservoir(self._gas)

    def _network(self, K, mdot=None):
        """ Create network connecting inlet/outlet to reactor. """
        if mdot is None:
            # Not very pythonic but ok in this case.
            mdot = lambda _: self._combustor.mass / self._tau

        pair = self._inlet, self._combustor
        inlet_mfc = ct.MassFlowController(*pair, mdot=mdot)

        pair = self._combustor, self._exhaust
        _ = ct.PressureController(*pair, master=inlet_mfc, K=K)

        self._sim = ct.ReactorNet([self._combustor])

    def set_states(self, 
            phi: float, 
            fuel: dict[str, float], 
            oxid: dict[str, float],
            T: Optional[float] = 300.0, 
            P: Optional[float] = ct.one_atm,
            basis: Optional[str] = "mole"
        ) -> None:
        """ Set source gas composition and create system.

        Parameters
        ----------
        phi: float
            Feed equivalence ratio used to set initial composition.
        fuel: dict[str, float]
            Dictionary of fuel composition compatible with Cantera.
        oxid: dict[str, float]
            Dictionary of oxidizer composition compatible with Cantera.
        T: Optional[float] = 300.0
            Feed gas temperature in kelvin.
        P: Optional[float] = ct.one_atm
            Feed gas pressure in pascal.
        basis: Optional[str] = "mole"
            Whether composition is in `mole` or `mass` fracitons.
        """
        self._gas.TP = T, P
        self._gas.set_equivalence_ratio(phi, fuel, oxid, basis=basis)
        self._inlet = ct.Reservoir(self._gas)
        self._equilibrate()

    def simulate_with_mass_flow(self,
            mdot: float,
            K: Optional[float] = 0.01
        ):
        """ Simulate combustor with imposed mass flow rate.
        
        Parameters
        ----------
        mdot : float
            Mass flow rate to pass throught the reactor [kg/s].
        K: Optional[float] = 0.01
            Valve constant used in pressure outlet.
        """
        self._states = ct.SolutionArray(self._gas, extra=["tau"])
        self._network(K, mdot)
        self._sim.set_initial_time(0.0)
        self._sim.advance_to_steady_state()
        self._states.append(self._combustor.thermo.state, tau=self._tau)

    def to_steady_state(self,
            tau_max: float,
            tau_min: Optional[float] = None,
            tau_dec: Optional[float] = 0.9,
            K: Optional[float] = 0.01
        ) -> None:
        """ Compute steady-state compositions for residence times.
        
        Parameters
        ----------
        tau_max: float
            Maximum residence time in seconds.
        tau_min: Optional[float] = None
            Minimum residence time in seconds. If not provided, it
            is set to one-thousandth of maximum value.
        tau_dec: Optional[float] = 0.9
            Decrease rate of residence time.
        K: Optional[float] = 0.01
            Valve constant used in pressure outlet.

        Raises
        ------
        AttributeError
            User did not call `set_states` method before computation.
        ValueError
            Invalid decrement rate for advancing residence time.
        """
        if self._combustor is None:
            raise AttributeError("Reactor has not been initialized.")

        if tau_dec >= 1.0:
            raise ValueError("Decrement rate must be below unity.")

        if tau_min is None:
            tau_min = 0.001 * tau_max

        self._tau = tau_max
        states = ct.SolutionArray(self._gas, extra=["tau"])

        self._network(K)

        while self._tau > tau_min:
            self._sim.set_initial_time(0.0)
            self._sim.advance_to_steady_state()
            states.append(self._combustor.thermo.state, tau=self._tau)
            self._tau *= tau_dec

        self._states = states

    def plot(self,
            tau_scale: float = 1000,
            species: list[str] = None,
            figsize: tuple[float, float] = (12, 5),
            xlim: tuple[float, float] = (None, None),
            ax1_ylim: tuple[float, float] = (None, None),
            ax2_ylim: tuple[float, float] = (None, None),
            ax3_ylim: tuple[float, float] = (None, None),
            loc: str | int = "best"
        ) -> Figure:
        """ Provides basic standard visualization of results.
        
        Parameters
        ----------
        tau_scale: float = 1000
            Scale to multiply residence time for better readability.
        species: list[str] = None
            List of species to plot. If none, it assumes partial 
            oxidation of methane is being studied.
        figsize: tuple[float, float] = (12, 5)
            Size of figure provided to matplotlib.
        xlim: tuple[float, float] = (None, None)
            Limits of x-axis (residence time) in both plots.
        ax1_ylim: tuple[float, float] = (None, None)
            Limits of y-axis in released energy in MW/m³.
        ax2_ylim: tuple[float, float] = (None, None)
            Limits of y-axis in combustion temperature.
        ax3_ylim: tuple[float, float] = (None, None)
            Limits of y-axis for species mole fractions.
        loc: str | int = "best"
            Placement of legend in species mole fractions plot.
        """
        names = self._states.species_names
        tau = tau_scale * self._states.tau
        qdot = self._states.heat_release_rate / 1e6
        T = self._states.T

        if species is None:
            species = ["CH4", "O2", "CO2", "H2O", "CO", "H2"]

        def plot_spec(spec, ax):
            """ Helper to plot species in axis. """
            if spec in names:
                idx = self._states.species_index(spec)
                X = self._states.X.T[idx]
                ax.loglog(tau, X, label=spec)

        plt.close("all")
        plt.style.use("seaborn-white")

        fig, (ax1, ax2) = plt.subplots(1, 2, figsize=figsize)
        ax3 = ax1.twinx()
    
        ax1.loglog(tau, qdot, ".-", color="C0")
        ax3.plot(tau, T, ".-", color="C1")

        for spec in species:
            plot_spec(spec, ax2)
        
        ax1.set_title(f"Temperature range {T.min():.1f}K - {T.max():.1f}K")
        ax2.set_title("Composition evolution")

        ax1.set_xlabel(f"Residence time [x{tau_scale} s]")
        ax2.set_xlabel(f"Residence time [x{tau_scale} s]")

        ax1.set_ylabel("Heat release rate [$MW\cdotp{}m^{-3}$]", color="C0")
        ax2.set_ylabel("Mole fraction")
        ax3.set_ylabel("Temperature [$K$]", color="C1",
                       rotation=270, labelpad=20)

        ax1.grid(linestyle=":")
        ax2.grid(linestyle=":")

        ax1.set_xlim(xlim)
        ax2.set_xlim(xlim)

        ax1.set_ylim(ax1_ylim)
        ax2.set_ylim(ax2_ylim)
        ax3.set_ylim(ax3_ylim)

        ax2.legend(loc=loc, fancybox=True, frameon=True, framealpha=1)

        fig.tight_layout()
        return fig

    @property
    def states(self):
        """ Provides access to computed states for post-processing. """
        if self._states is None:
            raise AttributeError("First call `to_steady_state`.")

        return self._states
