# -*- coding: utf-8 -*-
from typing import Optional
import cantera as ct


class HydrocarbonHeatingValue:
    """ Computes heating value of gas """
    def __init__(self, gas: ct.Solution) -> None:
        water = ct.Water()
        h_liq = self._water_enthalpy(water, 0.0)
        h_gas = self._water_enthalpy(water, 1.0)

        self._gas = gas
        self._hv = h_liq - h_gas

    def _water_enthalpy(self, water, Q):
        """ Compute enthalpy of water at given vapor quantity. """
        water.TQ = 298.15, Q
        return water.h

    def _heating_values(self,
            fuel: dict[str, float],
            oxid: dict[str, float],
            basis: Optional[str] = "mole"
        ) -> tuple[float, float]:
        """ Returns the LHV and HHV for the specified fuel.
        
        Implements a generalized version of Cantera's example:
        - cantera.org/examples/jupyter/thermo/heating_value.ipynb.html

        Note: there are some caveats in using this routine and the
        processing of values. In the reference example case, fuels
        are evaluated as pure substances. For real applications
        where this method is intended to be used, some species
        might be present in both fuel and oxidizer, *e.g.* carbon
        dioxide, water, and nitrogen. Thus, it has been chosen to
        compute `Y_fuel` based on the mass fraction of hydrocarbons
        in fuel only. This means that when using the heating values
        to compute required mass flow rates, one must take into
        account the contribution of those species in fuel and then
        scale according the full fuel composition.
        """
        # Set initial state with unit equivalence ratio.
        self._gas.TP = 298.15, ct.one_atm
        self._gas.set_equivalence_ratio(1.0, fuel, oxid, basis=basis)

        # Retrieve composition and enthalpy.
        Y_fuel = self.hydrocarbon_mass_fraction(self._gas, fuel)
        h1 = self._gas.enthalpy_mass

        # Set composition to complete combustion.
        X_products = self.complete_combustion_products(self._gas)
        self._gas.TPX = None, None, X_products
        
        # Retrieve composition and enthalpy.
        Y_H2O = self._gas["H2O"].Y[0]
        h2 = self._gas.enthalpy_mass

        # Apply definitions to compute heating values.
        lhv = -(h2 - h1) / (Y_fuel * 1000)
        hhv = lhv - (self._hv / 1000) * Y_H2O / Y_fuel

        return lhv, hhv

    @staticmethod
    def complete_combustion_products(
            gas: ct.Solution
        ) -> dict[str, float]:
        """ Compute expected complete combustion composition. """
        X_products = {
            "CO2": gas.elemental_mole_fraction("C"),
            "H2O": 0.5 * gas.elemental_mole_fraction("H"),
            "N2":  0.5 * gas.elemental_mole_fraction("N"),
        }
    
        # Notice the case convention for SPECIES / Element.
        if "AR" in gas.species_names:
            X_products["AR"] = gas.elemental_mole_fraction("Ar")

        # TODO add oxygen balance check here, there might be cases
        # where equivalence ratio has not been set to unit ?!?

        return X_products

    @staticmethod
    def select_hydrocarbons(
            gas: ct.Solution,
            fuel: dict[str, float]
        ) -> list[str]:
        """ Select hydrocarbons present in fuel dictionary. """
        keep = []

        for species in gas.species():
            if species.name not in fuel:
                continue

            atoms = sorted(species.composition.keys())
            if atoms == ["C", "H"] or atoms == ["H"]:
                keep.append(species.name)

        return keep

    @staticmethod
    def hydrocarbon_mass_fraction(
            gas: ct.Solution,
            fuel: dict[str, float]
        ) -> float:
        """ Fraction of hydrocarbons in mixture. """
        hc_species = HydrocarbonHeatingValue.select_hydrocarbons(gas, fuel)
        Y_fuel = sum(gas[hc_species].Y)
        return Y_fuel

    def heating_values(self,
            fuel: dict[str, float],
            oxid: dict[str, float],
            basis: Optional[str] = "mole"
        ) -> float:
        """ Compute corrected fuel heating value with provided method. """
        lhv, hhv = self._heating_values(fuel, oxid, basis=basis)

        self._gas.TPX = None, None, fuel
        corr = self.hydrocarbon_mass_fraction(self._gas, fuel)

        return corr * lhv, corr * hhv
